 
 
 // FAT - File - System ######################################################################################################
 // **************************************************************************************************************************
 // *   Das FAT - File - System / .c-File:                                                                                   *
 // *   =====================================                                                                                *
 // *   Dies ist das .c-File zum FAT-File-System. Fr eine genauere Beschreibung der Funktionen bitte                        *
 // *   das HEADER-File anschauen.                                                                                           *
 // *   (Hier befinden sich die Funktionen und Variablen.)                                                                   *                                                  *
 // *                                                                                                                        *
 // **************************************************************************************************************************

    #include "filefat.h" 

 //-------------------------------------------------------------------------------------------------   
 //      VARIABLEN
 //-------------------------------------------------------------------------------------------------
  //-----FAT-----
    unsigned char 	sector[512];				            // sektor puffer
    unsigned char	*bytesOfSec=&sector[0];		        // byte zeiger auf puffer:sector
    unsigned char 	bufferDirty=0;				          // zeigt an ob buffer dirty (wird von fat_setCluster gesetzt und von fat_loadSector gelesen)
    
    struct Fat{			        		                    // fat daten (1.cluster, root-dir, dir usw.)
    	unsigned char 		  secPerClust;		          // anzahl der sektoren pro cluster
    	unsigned long int 	currentSectorNr; 	        // aktuell geladener Sektor (in sector)  //beschleunigt wenn z.b 2* 512 byte puffer vorhanden, oder bei fat operationen im gleichen sektor		
    	unsigned long int 	dir; 		  	              // Direktory zeiger rootDir=='0' sonst(1.Cluster des dir; start auf root)	
    	unsigned long int  	rootDir;		              // Sektor(f16)/Cluster(f32) nr root directory
    	unsigned long int  	dataDirSec;		            // Sektor nr data area 
    	unsigned long int  	fatSec;	 		              // Sektor nr fat area
    	unsigned int  		  bytsPerSec;		            // bytes per sector	
    	unsigned char  		  fatType;		              // fat16 oder fat32 (16 oder 32)
    	} fat;
    
    struct File{						                        // datei infos	(32 bytes lang- 16 bei 512 bytesPerSec)												
    	unsigned char 		  name[13];		              // 0,10			datei Name.ext (8.3 = max 11)
    	unsigned char		    attrib;			              // 11,1			datei Attribut: 8=value name, 32=datei, 16=Verzeichniss, 15=linux kleingeschrieben eintrag
    	unsigned long int 	firstCluster;		          // 20,2 /26,2	datei 1.cluster hi,low(möglicherweise der einzige)	(4-byte)
    	unsigned long int	  length;			              // 28,4			datei Länge (4-byte)				
    	unsigned char 		  row;			                // reihe im sektor in der die datei infos stehen (reihe 0-15)
    	} file;	
    	 
    //-----File-----
    unsigned long int 	  seek;				              // zählt gelesene bytes (wird auch von ffseek genutzt)
    unsigned char		      cntSecOfClust;			      // zählt die gelesenen sektoren des clusters (2,4...max 128 siehe secPerCluster)
    unsigned char 		    file_flag=0;			        // 0== datei operation freigegeben, 1==read, 2==write, 3==append	
    unsigned long int 	  clusterChain[ccSize+1];		// puffert ccSize cluster der geöffneten datei 
    unsigned long int 	  *ccPointer=&clusterChain[0];	// cluster-chain-Pointer, zeigt auf stelle in puffer:clusterChain
    unsigned long int 	  currentCluster;			      // aktuell bearbeiteter cluster 
    unsigned int 		      cntOfBytes;			          // zäht geschriebene bytes eines sektors	
    	
    unsigned char*        kb_buffer[16];
            
  
    
  
 //-------------------------------------------------------------------------------------------------   
 //      FUNKTIONEN
 //-------------------------------------------------------------------------------------------------
 #pragma warn-                      //verhindert Warnung wegen evtl. Overflow bei Berechnungen
//*******************************************************************************************************************************
// öffnet EINE datei zum lesen/schreiben/anhängen. 
// datei optionen: lesen(1),schreiben/erstellen(2),anhängen(3).
//*******************************************************************************************************************************
unsigned char ffopen(unsigned char option,unsigned char name[]){		
  unsigned long int end, start,c;
  unsigned char	s;
    	
  // datei lesen *******************************************************************************	
  if(option=='r' && file_flag==0){					// option; dateioperation frei.		
	if(0==fat_loadFileDataFromDir(name)){				// läd datei infos aus dem direktory (wechselbar mit ffcd).	
	  bufferClusterChain(file.firstCluster);			// puffert die ersten ccSize(anzahl) cluster der datei.		
	  if(0==fat_loadSector(fat_clustToSec(file.firstCluster))){	// lät die ersten 512 bytes der datei auf puffer:sector.		
		seek=file.length;								// init von seek, noch keine bytes gelesen.
		cntSecOfClust=1;								// 1.sektor von cluster geladen.
		ccPointer++;									// zeiger++ weil 1. cluster schon geladen.	
		bytesOfSec=&sector[0];							// zeiger auf erstes byte, der datei-bytes.
		file_flag=1;									// 1==file read.				
		return(0);										// alles ok.									
		}					
	  }	
	}
	
  // datei schreiben ***************************************************************************	
  if(option=='w' && file_flag==0){				// option; dateioperation frei.				  
	if(1==fat_loadFileDataFromDir(name)){		// prüft ob datei im dir vorhanden, wenn nein, dann schreiben !
	  currentCluster=fat_getFreeCluster(2);		// holt neuen freien cluster.  		  
	  fat_setCluster(currentCluster,0x0fffffff);// fat16/32 cluster chain ende setzen.	
	  strcpy((char*)file.name,(char*)name);		// füllen des file struct, zum abschließenden schreiben.
	  file.firstCluster=currentCluster;			// 1. cluster der datei.
	  file.attrib=32;							// file.row wird in der funktion fat_getFreeRowOfDir geschrieben !! 
	  file.length=0;							// zählt geschriebene bytes.
	  cntSecOfClust=1;							// noch kein sektor geschrieben.
	  cntOfBytes=0;								// zählt geschriebene bytes eines sektors.
	  file_flag=2;								// schreiben flag.	
	  mmc_write_atAddr_start(fat_clustToSec(file.firstCluster));// teilt karte die aktuelle schreib adresse mit
	  return (0);								// alles OK !
	  }
	}

  // datei anhängen ***************************************************************************	
  if(option=='a' && file_flag==0){					// option; dateioperation frei.		  
	if(0==fat_loadFileDataFromDir(name)){			// prüft ob datei im dir vorhanden, wenn nicht, abbruch !
	  end = file.firstCluster;		// variablen, wegen verketteten clustern...
	  start = end;    		
	  while(!((end==0xfffffff&&fat.fatType==32)||(end==0xffff&&fat.fatType==16))){	// solange bis ende cluster chain		  
		start=end;		
		end=fat_getNextCluster(end);	
		}		  
		// TODO einfachere methode finden den gewünschten sektor zu bestimmen !!
	  end  = file.length>>9;												// end (/512) wird hier neu verwendet !
	  c	 = end/fat.secPerClust;				
	  s	 = (unsigned char)(end-(c*fat.secPerClust));	// s ist sektor des clusters.	  
	  fat_loadSector(fat_clustToSec(start)+s);			// läd sektor des letzten clusters der datei.
	  cntSecOfClust=s+1;								// sektor s des clusters zum schreiben geöffnet !	
	  bytesOfSec=&sector[0];							// zeigt auf erstes byte des geladenen sektors.
	  mmc_write_atAddr_start(fat_clustToSec(start)+s);	// teilt karte die adresse mit, ab der nun geschrieben wird.
	  // da man immer 512 bytes in folge schreiben muss, wird hier der inhalt der schon da ist nochmal auf die karte geschrieben,
	  // beim schreiben eines sektors (wo nur wenig geändert wurde) passiert nichts anderes...
	  for(cntOfBytes=0;cntOfBytes<(file.length%512);cntOfBytes++)mmc_write_byte(*bytesOfSec++);
	  file_flag=3;	  
	  return (0);											// alles OK !
	  }	
	}

  return(1);	//fehler?
}


//*******************************************************************************************************************************
// gibt ffopen wieder frei. regelt wenn zum schreiben/anhängen geöffnet war.
// da immer 512 bytes komplett zur karte geschrieben werden müssen, muss noch der rest des sektors,
// der nicht genutzt wurde (beim schreiben/anhängen) voll gemacht werden !!
// erst wenn datei geschlossen wird, wird wenn zum schreiben/erstellen geöffnet wird die datei informationen geschrieben,
// bis hier werden nur die daten geschrieben !
//*******************************************************************************************************************************
unsigned char ffclose(void){  
  unsigned char buffer[13];
  unsigned long int i,j;
  
  if(file_flag>1){								// war zum schreiben/anhängen geöffnet
	j=file.length%512;				// errechnet stelle ab der mit 0 überschrieben werden muss				
	while(j<512){									
	  mmc_write_byte(0x00);						// beschreibt ungenutzte bytes mit 0x00 	
	  j++;			
	  }			
	mmc_write_atAddr_end();						// abschluss, schreiben eines blocks

	if(file_flag==2)fat_makeFileEntry(file.name,file.attrib,file.firstCluster,file.length); // macht datei eintrag im aktuellen verzeichniss		

	if(file_flag==3){
	  i=file.length;			// muss gesichert werden, wird sonst von der karte geladen und verändert !	  
	  strcpy((char*)buffer,(char*)file.name);// muss gesichert werden, wird sonst von der karte geladen und verändert !
	  fat_loadFileDataFromDir(buffer);		// läd sektor, des datei eintrags, und läd daten auf struct file!	  	  	  
	  file.length=i;
	  fat_makeRowDataEntry(file.row,file.name,file.attrib,file.firstCluster,file.length);	 // macht eintrag im puffer
	  fat_writeSector(fat.currentSectorNr);									 // schreibt puffer auf karte
	  }
	}

  file_flag=0;		
  cntSecOfClust=0;
  seek=0;		
  cntOfBytes=0;
  return(0);	
}




//***************************************************************************************PhysFS****************************************
// wechselt verzeichniss. start immer im root Dir.
// MUSS in das direktory gewechselt werden, in dem die datei zum lesen/schreiben ist !
//*******************************************************************************************************************************
unsigned char ffcd(unsigned char name[]){	
	if(0==fat_cd(name))return(0);
	return(1);
}


//*******************************************************************************************************************************
// zeigt reihen eines clusters an, wird für ffls benötigt !
// es wird ab dem start sektor start_sec, der dazugehörige cluster angezeigt. geprüft wird ob es ein richtiger 
// eintrag in der reihe ist (nicht gelöscht, nicht frei usw). die sektoren des clusters werden nachgeladen.
// die dateien werden mit namen und datei größe angezeigt.
//*******************************************************************************************************************************
void lsRowsOfClust (unsigned long int start_sec){

  unsigned char row;						// reihen
  unsigned char sec=0;						// sektoren 
  
  do{
	fat_loadSector(start_sec + sec);			
	for(row=0;row<16;row++){								// geht durch reihen des sektors	
	  fat_loadRowOfSector(row);				
	  if((file.attrib==0x20||file.attrib==0x10) && (file.name[0]!=0xE5&&file.name[0]!=0x00)){	  
    //----Eigene Ausgabe--------
 	  puts1((char*)file.name);
	  putchar1(' ');
    ltoa(file.length,(char*)kb_buffer);
 	  puts1((char*)kb_buffer);			
//  	  printf("%lu",file.length);
    putchar1('\n');
    putchar1('\r');            
    //----Eigene Ausgabe--------
	  }
	}
	sec++;
	}while(sec<fat.secPerClust);
}


//*******************************************************************************************************************************
// zeigt inhalt eines direktory an.
// unterscheidung ob man sich im rootDir befindet nötig, weil bei fat16 im root dir eine bestimmt anzahl sektoren durchsucht
// werden müssen und bei fat32 ab einem start cluster ! ruft lsRowsOfClust auf um cluster/sektoren anzuzeigen.
//*******************************************************************************************************************************
unsigned char ffls(void){

  unsigned long int clust;											// cluster
  unsigned int s;													// fat16 root dir sektoren

  if(fat.dir==0 && fat.fatType==16){							// IM ROOTDIR.	fat16		
	for(s=0;s<(unsigned int)(fat.dataDirSec+2-fat.rootDir);s++){	// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).	
	  lsRowsOfClust(fat.rootDir+s);									// zeigt reihen eines root dir clust an
	  }
	}

  else {		
	if(fat.dir==0 && fat.fatType==32)clust=fat.rootDir;			// IM ROOTDIR.	fat32		
	else clust=fat.dir;											// NICHT ROOT DIR																
	while( !((clust==0xfffffff&&fat.fatType==32)||(clust==0xffff&&fat.fatType==16)) ){	// prüft ob weitere sektoren zum lesen da sind (fat32||fat16)	
	  lsRowsOfClust(fat_clustToSec(clust));												// zeigt reihen des clusters an
	  clust=fat_getNextCluster(clust);													// liest nächsten cluster des dir-eintrags
	  }
	}
  return(0);
}



#if smallFileSys 		// *******************************************************************************************************

// *******************************************************************************************************************************
// übergabe ist die position ab der gelesen werden soll (offset)!!
// (läd den absoluten sektor, des clusters auf puffer:sector, setzt den zeiger auf das gewünschte byte.)
// es wird die position des gewünschten clusters in der cluster chain errechnet(c) und der sektor(s) in dem cluster, dann wird 
// durch die cluster chain gegangen, bis der gewünschte cluster erreicht ist. der errechnete sektor in dem cluster wird geladen.
// unterscheidung noch ob man sich im ersten sektor des ersten clusters befindet (c=0 s=0 tmp=0).
// *******************************************************************************************************************************
unsigned char ffseek(unsigned long int offset){				
  unsigned long int tmp,c,i;
  unsigned char s;
  
  tmp = offset/fat.bytsPerSec;		// berechnungs temp variable (anzahl der sektoren, abgerundet)
	c	 = tmp/fat.secPerClust;			// cluster position in der chain	(abgerundet)			
	s	 = tmp-(c*fat.secPerClust);		// sektor des clusters in dem das byte mit position:offset ist	
	i	 = c+1;							// zählt die cluster ab, +1 weil vorher abgerundet	

  if(file_flag!=0){										// datei operation freigegeben
	seek-=offset;								// übersprungene bytes abziehen 
	bufferClusterChain(file.firstCluster);		// start 1. cluster der datei (while schleife sucht den richtigen von anfang an)

	while(--i){									// sucht den gewünschten cluster aus der fat 		
	  ccPointer++;								// zeigt auf clusteradresse im puffer:clusterChain				
	  if(ccPointer==&clusterChain[ccSize]) bufferClusterChain(*ccPointer);		// weiter laden der chain wenn gepufferte nicht ausreichen
	  }

    if(0==fat_loadSector( fat_clustToSec(*ccPointer) +s )){							// läd den sektor mit der byte position:offset
	  ccPointer++;																	// zeigt auf nächsten cluster, weil grade einen glesen
	  bytesOfSec=&sector[0];														// byte zeiger auf 1. stelle					
	  if(offset<fat.bytsPerSec)	bytesOfSec+=fat.bytsPerSec-(fat.bytsPerSec-offset);	// byte offset | kleiner 1. sektor
	  else 						bytesOfSec+=offset-tmp*fat.bytsPerSec;				// byte offset | größer  1. sektor
	  cntSecOfClust=s+1;															// sektor offset	(+1 weil einen grade geladen)			
	  return(0);		
	  }
  }

  return(1);	
}


//*******************************************************************************************************************************
// wechselt in das parent verzeichniss (ein verzeichniss zurück !)
// die variable fat.dir enthält den start cluster des direktory in dem man sich grade befindet, anhand diesem,
// kann der "." bzw ".." eintrag im ersten sektor des direktory ausgelesen und das parent direktory bestimmt werden.
//*******************************************************************************************************************************
unsigned char ffcdLower(void){

	if(fat.dir==0)return(1);							// im root dir, man kann nicht höher !
	
	fat_loadSector(fat_clustToSec(fat.dir));			// läd 1. sektor des aktuellen direktory.
	fat_loadRowOfSector(1);								// ".." eintrag (parent dir) ist 0 wenn parent == root

	if(file.firstCluster==0)fat.dir=0;					// zum root dir
	else fat.dir=file.firstCluster;						// nicht zum root dir
	
	return(0);
}


//*******************************************************************************************************************************
// löscht datei aus aktuellem verzeichniss, wenn es die datei gibt.
// geht nur wenn keine schreib/lese/anhäng vorgänge statfinden. löscht nur dateien !
//*******************************************************************************************************************************
unsigned char ffrm(unsigned char name[]){
  unsigned long int parent;
	unsigned long int own;
	unsigned long int clustsOfDir;		// um durch die cluster chain eines dirs zu gehen.
	unsigned char i=0;    
	
  if(0==fat_loadFileDataFromDir(name) && file_flag==0){	// datei gibt es nicht oder file operation, abbruch.

	if(file.attrib!=0x10){			// ist datei.	
	  sector[file.row<<5]=0xE5;		// datei gelöscht markieren (file.row*32, damit man auf reihen anfang kommt)
	  if((file.row-1)>=0){			// gibt es einen eintrag davor ?
	  if(sector[(file.row<<5)-21]==0x0f)sector[(file.row<<5)-32]=0xE5;	// langer datei eintag auch gelöscht.
		}	
	  bufferDirty=1;								// eintrag in puffer gemacht. 	  
	  fat_delClusterChain(file.firstCluster);		// löscht die zu der datei gehörige cluster-chain aus der fat.
	  return(0);
	  }
 
  //TODO noch nicht optimal. zu viele schreib vorgänge beim löschen von datei einträgen. bis zu 16/sektor !
  else{									// ist ordner, rekursiv löschen !!
	sector[file.row<<5]=0xE5;		// löscht dir eintrag
	if((file.row-1)>=0){			// gibt es einen eintrag davor (langer datei eintrag)?
	  if(sector[(file.row<<5)-21]==0x0f)sector[(file.row<<5)-32]=0xE5;	// langer datei eintag auch gelöscht
	  }
	bufferDirty=1;					// puffer eintrag gemacht
		
	parent=fat.dir;										// der ordner in dem der zu löschende ordner ist.
	own=file.firstCluster;								// der 1. cluster des ordners der zu löschen ist.
	clustsOfDir=file.firstCluster;						// die "cluster" des zu löschenden ordners
	cntSecOfClust=0;									// 0 sektoren bearbeitet

	do{													// geht durch cluster des dirs							
	  fat_loadSector(fat_clustToSec(clustsOfDir));		// sektor des dirs laden
	  do{												// geht durch sektoren des clusters
		do{	  											// geht durch reihen des sektors
		  fat_loadRowOfSector(i);

		  if(file.attrib!=0x10 && file.attrib!=0x00 && file.name[0]!=0xE5){	// ist kein ordner,noch nicht gelöscht, kein freier eintrag
			sector[i<<5]=0xE5;									// erster eintrag der reihe als gelöscht markiert		
			bufferDirty=1;										// puffer eintrag gemacht
			if(file.attrib==0x20){								// ist datei!
			  fat_delClusterChain(file.firstCluster);						// ist datei, dann cluster-chain der datei löschen
			  fat_loadSector(fat_clustToSec(clustsOfDir)+cntSecOfClust);	// sektor neu laden, weil löschen der chain, den puffer nutzt.			
			  }
			}

		  if(file.attrib==0x10&&file.name[0]=='.'){				// "." oder ".." eintrag erkannt, löschen !
			sector[i<<5]=0xE5;									// eintrag als gelöscht markiert		
			bufferDirty=1;										// puffer eintrag gemacht		  
			}

		  if(file.attrib==0x10&&file.name[0]!='.'&&file.name[0]!=0xE5){	// ordner erkannt. löschen und rein wechseln !
			sector[i<<5]=0xE5;									// dir eintrag als gelöscht markiert		
			bufferDirty=1;										// puffer eintrag gemacht		  
			fat_loadSector(fat_clustToSec(file.firstCluster));	// sektor des dirs laden		  
			clustsOfDir=file.firstCluster;						// eigenes dir ist file.firstCluster
			own=file.firstCluster;								// eigener start cluster/dir
			fat_loadRowOfSector(1);								// ".." laden um parent zu bestimmen
			parent=file.firstCluster;							// parent sichern.		  
			cntSecOfClust=0;									// init von gelesenen sektoren und reihen !
			i=0;
			continue;
			}

		  if(file.name[0]==0x00){								// ende des dirs erreicht, wenn nicht voll !!				  
			if(parent==fat.dir){								// erfolgreich alles gelöscht
			  fat_delClusterChain(own);							// cluster chain des ordners löschen
			  return(0);
			  }
			fat_delClusterChain(own);							// cluster chain des ordners löschen
			clustsOfDir=parent;									// parent ist jetzt wieder arbeisverzeichniss.		 
			own=parent;											// arbeitsverzeichniss setzten
			fat_loadSector(fat_clustToSec(own));				// sektor des dirs laden
			fat_loadRowOfSector(1);								// ".." laden um parent zu bestimmen
			parent=file.firstCluster;		  		  			// parent sichern.
			cntSecOfClust=0;									// init von gelesenen sektoren und reihen !
			i=0;
			continue;
			}
	  
		  i++;
		  }while(i<16);													// geht durch reihen des sektors.
	
		i=0;															// neuer sektor -> reihen von vorne.
		cntSecOfClust++;	  											
		fat_loadSector(fat_clustToSec(clustsOfDir)+cntSecOfClust);		// läd sektoren des clusters nach
		}while(cntSecOfClust<fat.secPerClust);							// geht durch sektoren des clusters.

	  cntSecOfClust=0;													// neuer cluster -> sektoren von vorne.
	  clustsOfDir=fat_getNextCluster(clustsOfDir);						// sucht neuen cluster der cluster-chain.	  
	  }while(!((clustsOfDir==0xfffffff&&fat.fatType==32)||(clustsOfDir==0xffff&&fat.fatType==16)));		// geht durch cluster des dirs.
	  fat_delClusterChain(own);				// hier landet man, wenn der ordner voll war !!
	  }

  }

  return(1);				// fehler
}


// *******************************************************************************************************************************
// erstellt einen dir eintrag im aktuellen verzeichniss (geht nur wenn keine lese/schreib usw. vorgänge).
// prüft ob es den den dir-namen:name schon gibt, dann wird nichts angelegt.
// wenn ok, dann wird ein freier cluster gesucht, als ende markiert, der eintrag ins dir geschrieben.
// dann wird der cluster des dirs aufbereitet. der erste sektor des clusters enthält den "." und ".." eintrag.
// der "." hat den 1. cluster des eigenen dirs. der ".." eintrag ist der 1. cluster des parent dirs.
// ein dir wird immer mit 0x00 initialisiert ! also alle sektoren des clsters ( bis auf . und .. einträge)!
// *******************************************************************************************************************************
unsigned char ffmkdir(unsigned char name[]){
  unsigned char i;
  char* TEMP = ".           ";
  char* TEMP1= "..          ";
  
 // unsigned long int sec=fat.currentSectorNr;

  if(0==fat_loadFileDataFromDir(name)||file_flag!=0)return (1);	// prüft ob dirname im dir schon vorhanden, wenn ja, abbruch !			

  currentCluster=fat_getFreeCluster(0);				// holt neuen freien cluster.  		  
  fat_setCluster(currentCluster,0x0fffffff);		// fat16/32 cluster chain ende setzen.	  

  fat_makeFileEntry(name,0x10,currentCluster,0); 	// macht dir eintrag im aktuellen verzeichniss	

  // aufbereiten des neuen dir clusters !
  fat_markSector00();					// löschen des sektors (nur im puffer).
  //fat_makeRowDataEntry(0,(unsigned char *)strings(PSTR(".           ")),0x10,currentCluster,0);	// macht "." eintrag des dirs
  fat_makeRowDataEntry(0,TEMP,0x10,currentCluster,0);		// macht "." eintrag des dirs
  fat_makeRowDataEntry(1,TEMP1,0x10,fat.dir,0);			// macht ".." eintrag des dirs
  //fat_makeRowDataEntry(1,(unsigned char *)strings(PSTR("..          ")),0x10,fat.dir,0);			// macht ".." eintrag des dirs
  fat_writeSector(fat_clustToSec(currentCluster));					// schreibt einträge auf karte !

  fat_markSector00();					// löschen des sektors (nur im puffer).  
  
  for(i=1;i<fat.secPerClust;i++){		// ein dir cluster muss mit 0x00 initialisiert werden !
	fat_writeSector(fat_clustToSec(currentCluster)+i);	// löschen des cluster (überschreibt mit 0x00), wichtig bei ffls!
	}	

  //fat_loadSector(sec);
 
  return(0);

}


#endif				// *******************************************************************************************************


//*******************************************************************************************************************************
// liest ein byte aus der geöffneten datei.
// läd sektoren der datei nach. wenn ein cluster mehrere sektoren hat, dann werden diese nacheinander gelesen.
// puffert die cluster chain der datei zwischen.
// gibt bytes der datei zurück (immer 1 byte ). 
//*******************************************************************************************************************************
unsigned char ffread(void){
  
  if(file_flag!=1)return(1);						// datei zum lesen geöffnet	(nur dann lesen möglich)										
	if(bytesOfSec==&sector[512]){					// einen sektor der datei gelesen 															
	  if(cntSecOfClust<fat.secPerClust){			// nächsten sektor des clusters lesen, wenn mehrere sektoren/cluster
		fat_loadSector(fat.currentSectorNr+1);		// nächsten sektor auf puffer:sector geladen				
		bytesOfSec=&sector[0];						// zurücksetzten von byte zeiger (gelesene bytes des sektors)
		cntSecOfClust++;							// gelesene sektoren des clusters hochzählen
		}
	  else {		
		if(ccPointer==&clusterChain[ccSize]) bufferClusterChain(*ccPointer);	// weiter laden der chain, wenn nötig					
		if(0==fat_loadSector( fat_clustToSec(*ccPointer) )) {		// läd neuen sektor auf puffer:sector(absoluter sektor)					
		  ccPointer++;								// cluster gelesen, auf nächsten zeigen
		  bytesOfSec=&sector[0];					// byte zeiger auf erstes byte setzten
		  cntSecOfClust=1;							// 1. sektor des clusters gelesen	
		  }
		}		
	}		
  return *bytesOfSec++;							// rückgabe, bytes des sektors				
 }




//*******************************************************************************************************************************
// schreibt ein byte der geöffneten datei, direkt auf die karte. wenn ein neuer cluster benötigt wird, wird dieser gesucht und
// verkettet. es werden alle sektoren eines clusters beschrieben.
// die karte bekommt die schreib adresse mitgeteilt, danach müssen 512 bytes in folge kommen, schreib aktion wird beendet,
// dann kann neuer sektor beschrieben werden (beschreibt sektoren von clustern). ist ein cluster beschrieben,
// wird ein neuer gesucht, verkettet und schreiben geht weiter.
//*******************************************************************************************************************************
unsigned char ffwrite(unsigned char c){
	unsigned long int clusterOld;
	if(file_flag>1){											// prüft ob datei operation w oder a.
	mmc_write_byte(c);											// c auf karte schreiben
	cntOfBytes++;												// zähler für geschriebene bytes eines sekors
	file.length++;												// geschriebene bytes erhöhen		

	if(cntOfBytes==512){								// SEKTOR GESCHRIEBEN ( 2 möglichkeiten ab hier !)	
	  mmc_write_atAddr_end();									// abschluss schreiben eines sektors
	  cntOfBytes=0;												// init des zählers

	  if(cntSecOfClust<fat.secPerClust){				// 1. MEHRERE SEKTOREN pro CLUSTER				
		mmc_write_atAddr_start(fat_clustToSec(currentCluster)+cntSecOfClust); // teilt karte neue adresse mit				
		cntSecOfClust++;									// geschriebene sektoren des clusters erhöhen	
		return (0);
		}
	  else{												// 2. NEUER CLUSTER NÖTIG (alle sektoren des clusters geschrieben)			
		clusterOld=currentCluster;		// tmp variable zum verketten der cluster				
		cntSecOfClust=1;									
		currentCluster=fat_getFreeCluster(clusterOld);		// holt neuen freien cluster		
		fat_setCluster(clusterOld,currentCluster);			// cluter chain verketten
		fat_setCluster(currentCluster,0xfffffff);			// cluster chain ende		
		mmc_write_atAddr_start(fat_clustToSec(currentCluster));	// teilt karte neue adresse mit				
		return (0);
		}				
	  }		

 	return (0);
	}

  return (1);
}


//*******************************************************************************************************************************
// puffert cluster-chain einer datei die zum lesen geöffnet ist.
// läd die cluster-chain der datei und puffert diese auf: clusterChain	(immer ccSize cluster). beschleunigt lesen, weil nicht
// immer 512 bytes der fat gelesen werden müssen,  um den nächsten cluster zu bestimmen(braucht mehr ram).
// man müsste 512 bytes lesen um an die nächsten 512 bytes daten zu kommen, macht 1KB für 1/2KB daten = overhead von 50 prozent.
//*******************************************************************************************************************************
unsigned char bufferClusterChain(unsigned long int firstCluster){
	
  unsigned char i=ccSize;			// sooft wie ccSize groß puffern
  unsigned char k=0;		
					
  ccPointer=clusterChain	;		// init des zeigers
  clusterChain[0]=firstCluster;		// letzter cluster ist der start zum suchen
	
  do{		
	clusterChain[k+1]=fat_getNextCluster(clusterChain[k]);				// liest cluster aus fat und puffer diese		
	if( (clusterChain[k+1]==0xfffffff && fat.fatType==32) || (clusterChain[k+1]==0xffff && fat.fatType==16) )return(0);// cluster chain ende zeichen  
	k++;					
	}while(--i);
	
  return(1);
}




//***************************************************************************************************************
// markiert sector komplett mit 0x00 (nur den gepufferten)
// wenn neuer direktory cluster verkettet wird, muss er mit 0x00 initialisiert werden (alle felder)
//***************************************************************************************************************
void fat_markSector00(void){
	  unsigned int i=511;	
	  	
  fat.currentSectorNr=0;	// geladener sektor wird hier geändert !
  do{
	sector[i]=0x00;		//schreibt puffer sector voll mit 0x00==leer
	}while(--i);		
  sector[0]=0x00;		// noch die stelle 0 überschreiben		
}


// ***************************************************************************************************************
// umrechnung cluster auf 1.sektor des clusters (möglicherweise mehrere sektoren/cluster !
// ***************************************************************************************************************
unsigned long int fat_clustToSec(unsigned long int clust){

	return fat.dataDirSec+2+((clust-2) * fat.secPerClust);		// errechnet den 1. sektor der sektoren des clusters	
}


//***************************************************************************************************************
// läd sektor:sec auf puffer:sector zum bearbeiten im ram !
// setzt currentSectorNr auf richtigen wert (also den sektor der gepuffert ist). es wird geprüft
// ob der gepufferte sektor geändert wurde, wenn ja muss erst geschrieben werden, um diese daten nicht zu verlieren !
//***************************************************************************************************************
unsigned char fat_loadSector(unsigned long int sec){
	
  if(sec!=fat.currentSectorNr){					// nachladen nötig		
	if(bufferDirty==1)fat_writeSector(fat.currentSectorNr);	// puffer diry, also vorher schreiben		  		  
	if(0==mmc_read_sector(sec,sector)){			// neuen sektor laden					
	  fat.currentSectorNr=sec;					// aktualisiert sektor nummer (nummer des gepufferten sektors)
	  return(0);
	  }
	}	
		
  else return(0);								// alles ok, daten sind schon da (sec==fat.currentSectorNr)	

  return(1);									// fehler	
}


//***************************************************************************************************************
// schreibt sektor nummer:sec auf die karte (puffer:sector) !!
// setzt bufferFlag=0 da puffer nicht dirty sein kann nach schreiben !
//***************************************************************************************************************
unsigned char fat_writeSector(unsigned long int sec){	

  if(0==mmc_write_sector(sec,sector)){
	bufferDirty=0;	// buffer kann nicht dirty sein weil grade geschrieben		
	return(0);		// schreiben von sektor puffer											
	}

  return (1);
}






// datei lesen funktionen:

// fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_loadFileDataFromDir -> fat_cd   "daten chain"

//***************************************************************************************************************
// läd die reihe:row des gepufferten sektors auf das struct:file. dort stehen dann
// alle wichgigen daten wie: 1.cluster,länge bei dateien, name des eintrags, reihen nummer (im sektor), attribut use...
//***************************************************************************************************************
unsigned char fat_loadRowOfSector(unsigned int row){
  void *firstCluster;
  void *length;
	unsigned char i;						// zähler im sektor(32byte zähler).
	row=row<<5;								// multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 usw)	.

	bytesOfSec=&sector[row];				// byte zeiger auf puffer:sector.
	
	firstCluster=&file.firstCluster;	// void pointer auf datei erster cluster,zum schreiben von einzel bytes auf 4 byte variable.
	length=&file.length; 				// void pointer auf datei länge, zum schreiben von einzel bytes auf 4 byte variable.

	for(i=0;i<11;i++)  file.name[i]=*bytesOfSec++;		// datei name, ersten 10 bytes vom 32 byte eintrag.	  
	  
	file.attrib=*bytesOfSec;							// datei attribut, byte 11.
	
	*(unsigned char*)firstCluster++=sector[row+26];		// datei erster cluster	, byte von klein nach hoch: 26,27,20,21	.
	*(unsigned char*)firstCluster++=sector[row+27];		// hier nicht mit bytesOfSec pointer, weil man hin und her springen,
	*(unsigned char*)firstCluster++=sector[row+20];		// müsste..
	*(unsigned char*)firstCluster++=sector[row+21];

	bytesOfSec+=17;

	*(unsigned char*)length++=*bytesOfSec++;			// datei länge, bytes 28,29,30,31.
	*(unsigned char*)length++=*bytesOfSec++;
	*(unsigned char*)length++=*bytesOfSec++;
	*(unsigned char*)length++=*bytesOfSec++;

	return(0);				
}


//***************************************************************************************************************
// geht reihen weise durch sektoren des clusters mit dem startsektor:sec, und sucht nach der datei mit dem 
// namen:name. es werden die einzelnen sektoren nachgeladen auf puffer:sector vor dem bearbeiten.
// wird die datei in dem cluster gefunden ist return 0 , sonst return1.
//***************************************************************************************************************
unsigned char fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]){
	
  unsigned char r;  
  unsigned char s=0;

  do{											// sektoren des clusters prüfen
	if(0==fat_loadSector(sec+s)){				// läd den sektor:sec auf den puffer:sector		
	  for(r=0;r<16;r++){						// zählt zeilennummer (16(zeilen) * 32(spalten) == 512 bytes des sektors)
		fat_loadRowOfSector(r);					// zeile 0-15 auf struct:file laden 
		if(0==strncmp((char*)file.name,(char*)name,10)){	// zeile r ist gesuchte
		  file.row=r;										// zeile sichern.
		  return(0);			
		  }
		}
	}
	s++;
	}while(s<fat.secPerClust);		// geht durch sektoren des clusters

	return (1);									// fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// wenn dir == 0 dann wird das root direktory durchsucht, wenn nicht wird der ordner cluster-chain gefolgt, um
// die datei zu finden. es wird das komplette directory in dem man sich befindet durchsucht.
// bei fat16 wird der rootDir berreich durchsucht, bei fat32 die cluster chain des rootDir.
//***************************************************************************************************************
unsigned char fat_loadFileDataFromDir(unsigned char name[]){
  unsigned long int i;
  unsigned int d;  

  if(fat.dir==0 && fat.fatType==16){									// IM ROOTDIR.	fat16		
	for(d=0;d<(unsigned int)(fat.dataDirSec+2-fat.rootDir);d++){			// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_loadFileDataFromCluster(fat.rootDir+d,name))return(0);		// sucht die datei, wenn da, läd daten (1.cluster usw)															
	  }
	}		
	
  else {		
	
	if(fat.dir==0 && fat.fatType==32)i=fat.rootDir;						// IM ROOTDIR.	fat32		
	else i=fat.dir;																												
	while(!((i==0xfffffff&&fat.fatType==32)||(i==0xffff&&fat.fatType==16))){	// prüft ob weitere sektoren zum lesen da sind (fat32||fat16)					=
	  if(0==fat_loadFileDataFromCluster(fat_clustToSec(i),name))return(0);		// lät die daten der datei auf struct:file. datei gefunden (umrechnung auf absoluten sektor)																									
	  i=fat_getNextCluster(i);													// liest nächsten cluster des dir-eintrags (unterverzeichniss größer 16 einträge)								
	  }
	}

  return(1);																	// datei/verzeichniss nicht gefunden
}


//***************************************************************************************************************
// start immer im root Dir. start in root dir (dir==0).
// es MUSS in das direktory gewechselt werden, in dem die datei zum lesen/anhängen ist (außer root, da startet mann)!
//***************************************************************************************************************
unsigned char fat_cd(unsigned char name[]){

  if(name[0]==0){								// ZUM ROOTDIR FAT16/32
	fat.dir=0;										// root dir 
	return(0);
	}

  if(0==fat_loadFileDataFromDir(name)){			// NICHT ROOTDIR	(fat16/32)
	fat.dir=file.firstCluster;						// zeigt auf 1.cluster des dir	(fat16/32)	
	return(0);
	}

  return(1);									// dir nicht gewechselt (nicht da?!?) !!
}



// datei anlegen funktionen :

//fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector -> eintrag gemacht !!

// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im cluster mit dem startsektor:secStart.
// wird dort kein freier eintrag gefunden ist return (1).
// wird ein freier eintrag gefunden, ist die position der freien reihe auf file.row abzulesen und return (0).
// der sektor mit der freien reihe ist auf dem puffer:sector gepuffert.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfCluster(unsigned long secStart){
  
  unsigned int 	i;										// 1. zeichen einer reihe, eine sektors, eines clusters.
  unsigned char s=0;									// sektoren des clusters.
  
  do{
	file.row=0;											// neuer sektor(oder 1.sektor), reihen von vorne.
	if(0==fat_loadSector(secStart+s)){					// läd sektor auf puffer:buffer.dSector, setzt buffer.currentDatSector.
	  for(i=0;i<512;i=i+32){							// zählt durch zeilen (0-15).									  
		if(sector[i]==0x00||sector[i]==0xE5)return(0);	// prüft auf freihen eintrag (leer oder gelöscht == OK!).
		file.row++;										// zählt reihe hoch (nächste reihe im sektor).	
		}
	  }	
	s++;												// sektoren des clusters ++ weil einen geprüft.	
	}while(s<fat.secPerClust);							// geht die sektoren des clusters durch (möglicherweise auch nur 1. sektor).
  return (1);											// nicht gefunden in diesem cluster (== nicht OK!).
}


// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im directory mit dem startcluster:dir.
// geht die cluster chain des direktories durch. dabei werden auch alle sektoren der cluster geprüft.
// wird dort kein freier eintrag gefunden, wird ein neuer leerer cluster gesucht, verkettet und der 
// 1. sektor des freien clusters geladen. die reihe wird auf den ersten eintrag gesetzt, da frei.
// anhand der reihe kann man nun den direktory eintrag vornehmen, und auf die karte schreiben.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfDir(unsigned long int dir){
  unsigned char i;
  unsigned long int start;
  start=dir;   
 
  // solange bis ende cluster chain.
  while( !((dir==0xfffffff&&fat.fatType==32)||(dir==0xffff&&fat.fatType==16)) ){		  
	if(0==fat_getFreeRowOfCluster(fat_clustToSec(dir)))return(0);	// freien eintrag in clustern, des dir gefunden !!
	start=dir;		
	dir=fat_getNextCluster(dir);	
	}										// wenn aus schleife raus, kein freier eintrag da -> neuer cluster nötig.

  dir=fat_getFreeCluster(start);			// muss neuen freien cluster suchen 	(benutzt puffer:sector).  
  fat_setCluster(start,dir);				// cluster-chain mit neuem verketten	(benutzt puffer:sector).	
  fat_setCluster(dir,0x0fffffff);			// cluster-chain ende markieren	  		(benutzt puffer:sector).    
  fat_loadSector(fat_clustToSec(dir));		// läd neuen cluster (wenn bufferFlag==1 schreibt vorher alten cluster) !!
  fat_markSector00();						// löschen des sektors (nur im puffer) (benutzt puffer:sector).

  
  for(i=1;i<fat.secPerClust;i++){			// i=1 weil der 1. eintrag eh noch geschrieben wird... wäre hier doppelt.
	fat_writeSector(fat.currentSectorNr+i);	// löschen des cluster (überschreibt mit 0x00), wichtig bei ffls,
	}										// weil wenn cluster nicht mit 0x00 überschrieben wird, werden mögilcherwies
											// alte daten ausgegeben !!!!
  file.row=0;								// erste reihe frei, weil grad neuen cluster verkettet.

  return(0);						// keinen freien platz in clustern, des dir gefunden, aber neuen cluster verkettet.
}


//***************************************************************************************************************
// erstellt 32 byte eintrag einer datei, oder verzeichnisses im puffer:sector.
// erstellt eintrag in reihe:row, mit namen:name usw... !!  
// muss noch auf die karte geschrieben werden ! nicht optimiert auf geschwindigkeit.
//***************************************************************************************************************
unsigned char fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){
  unsigned char i;								//byte zähler in reihe von sektor (32byte eintrag)
  void *vLength;
  bufferDirty=1;								// puffer beschrieben, also neue daten darin(vor lesen muss geschrieben werden)
  row=row<<5;									//multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 ... zeile 15=480)	

  bytesOfSec=&sector[row];						// init des zeigers
				
  for(i=0;i<32;i++){													

  if(i<11)*bytesOfSec=name[i];					//namen schreiben (format beachten!!) (10 byte)

	if(i==11)*bytesOfSec=attrib;				//attrib schreiben (1 byte)

	if(i>11&&i<20)*bytesOfSec=0x01;				//nulle nicht nötige felder !
	if(i>21&&i<26)*bytesOfSec=0x01;				//nulle nicht nötige felder !

	if(i==20){									// 1. cluster der datei byte 20/21 und 26/27 = 4 bytes
	  *bytesOfSec++=(cluster&0x00ff0000)>>16;	// 1. low							
	  *bytesOfSec=(cluster&0xff000000)>>24;		// high byte																
	  i++;
	  }	
		
	if(i==26){												
	  *bytesOfSec++=(cluster&0x000000ff);		// low byte														
	  *bytesOfSec=(cluster&0x0000ff00)>>8;		// 2. low						
	  i++;
	  }	
		
	if(i==28){									//datei länge (4 bytes)					
	  vLength=&length;
	  *bytesOfSec++ =*(unsigned char*)vLength++;	// low						
	  *bytesOfSec++ =*(unsigned char*)vLength++;	//	1. hi			
	  *bytesOfSec++ =*(unsigned char*)vLength++;	//	2. hi 				
	  *bytesOfSec =*(unsigned char*)vLength;		//	hi		
	  return (0);								// alles ok, eintrag angelegt !!		
	  }
	bytesOfSec++;		
	}
		
  return (1);		//fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// macht den datei eintrag im jetzigen verzeichniss (fat.dir).
// file.row enthält die reihen nummer des leeren eintrags, der vorher gesucht wurde, auf puffer:sector ist der gewünschte
// sektor gepuffert. für fat16 im root dir muss andere funktion genutzt werden, als fat_getFreeRowOfDir (durchsucht nur dirs).
// fat.rootDir enthält bei fat32 den start cluster des directory, bei fat16 den 1. sektor des rootDir bereichs!
//***************************************************************************************************************
unsigned char fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){
  
  unsigned int s;														// zähler für root dir sektoren fat16
 
  if(fat.dir==0&&fat.fatType==32)fat_getFreeRowOfDir(fat.rootDir);	// IM ROOT DIR (fat32)

  else if(fat.dir==0 && fat.fatType==16){							// IM ROOT DIR (fat16)		
	for(s=0;s<(unsigned int)(fat.dataDirSec+2-fat.rootDir);s++){		// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_getFreeRowOfCluster(fat.rootDir+s))break;				// geht durch sektoren des root dir.
	  }
	}  

  else fat_getFreeRowOfDir(fat.dir);								// NICHT ROOT DIR
		
  fat_makeRowDataEntry(file.row,name,attrib,cluster,length);			// macht file eintrag im puffer	
  fat_writeSector(fat.currentSectorNr);									// schreibt file daten auf karte		  
  
  return(0);
	  
}







// fat funktionen:

//***************************************************************************************************************
// sucht nötige folge Cluster aus der fat ! 
// erster daten cluster = 2, ende einer cluster chain 0xFFFF (fat16) oder 0xFFFFFFF, 0xFFFFFF8 (fat32), 
// stelle des clusters in der fat, hat als wert, den nächsten cluster. (1:1 gemapt)!
//***************************************************************************************************************
unsigned long int fat_getNextCluster(unsigned long int oneCluster){	
  unsigned long int i,j;   
  void *vi;
  
  // FAT 16**************FAT 16	
  if(fat.fatType==16){					
												// i=oneCluster/256
	i=oneCluster>>8;;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-256*i)*2 == 2*oneCluster-512*i
	j=(oneCluster<<1)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)
			
	if(0==fat_loadSector(i+fat.fatSec)){		// ob neu laden nötig wird von fat_loadSector geprüft
	  i=0;  
	  vi=&i;								// zeiger auf i
	  bytesOfSec=&sector[j];	  				// zeiger auf puffer
	  *(unsigned char*)vi++=*bytesOfSec++;		// setzen low byte von i, aus puffer	  
	  *(unsigned char*)vi++=*bytesOfSec++;		// setzen von höherem byte in i, aus puffer
	  return i;									// gibt neuen cluster zurück (oder 0xffff)
	  }	
	}

  // FAT 32**************FAT 32	
  else{
												// i=oneCluster/128
	i=oneCluster>>7;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-128*i)*4 == oneCluster*4-512*i
	j=(oneCluster<<2)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)		
		
	if(0==fat_loadSector(i+fat.fatSec)){		// ob neu laden nötig wird von fat_loadSector geprüft
	  vi=&i;
	  bytesOfSec=&sector[j];
	  *(unsigned char*)vi++=*bytesOfSec++;
	  *(unsigned char*)vi++=*bytesOfSec++;	 
	  *(unsigned char*)vi++=*bytesOfSec++;
	  *(unsigned char*)vi++=*bytesOfSec++;	 
	  return i;
	  }	
  }		

  return(0);									// neuladen des fat sektors, in dem oneCluster ist nötig !!		
}


//***************************************************************************************************************
// sucht freien cluster in der fat ab dem offsetCluster (also offsetCluster+1).
//***************************************************************************************************************
unsigned long int fat_getFreeCluster(unsigned long int offsetCluster){
	
  do{											
	offsetCluster++;							// cluster nummer	
	}while(fat_getNextCluster(offsetCluster));	// freier cluster gefunden (return 0, weil inhalt des clusters 0)

  return offsetCluster;							
}


//***************************************************************************************************************
// setzt den cluster inhalt. errechnet den sektor der fat in dem cluster ist, errechnet das low byte von
// cluster und setzt dann byteweise den inhalt:content.
// prüft ob buffer dirty (zu setztender cluster nicht in jetzt gepuffertem).
// prüfung erfolgt in fat_loadSector, dann wird alter vorher geschrieben, sonst gehen dort daten verloren !!
//***************************************************************************************************************
unsigned char fat_setCluster(unsigned long int cluster, unsigned long int content){  
  unsigned long int i,j;
  void *vc;
	// FAT 16**************FAT 16	
	if(fat.fatType==16){					
													//	i=cluster/256
		i=cluster>>8;				// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-256*i)*2 == 2*cluster-512*i
		j=(cluster<<1)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat 
		
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)
//  			printf("\n\rfat_setClust: %lu-> %lu",cluster,content);
			bytesOfSec=&sector[j];					// init des zeigers auf low byte
			vc=&content;						// init des zeigers auf content
			*bytesOfSec++=*(unsigned char*)vc++;	// setzen von 2 byte..
			*bytesOfSec++=*(unsigned char*)vc++;
			bufferDirty=1;							// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0); 											
			}	
		}

	// FAT 32**************FAT 32	
	else{
													 // i=cluster/128
		i=cluster>>7;				// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-128*i)*4 == cluster*4-512*i
		j=(cluster<<2)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat
				
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)			
//  			printf("\n\rfat_setClust: %lu-> %lu",cluster,content);
			bytesOfSec=&sector[j];					// init des zeigers auf low byte	
			vc=&content;						// init des zeigers auf content
			*bytesOfSec++=*(unsigned char*)vc++;	// setzen von 4 byte....
			*bytesOfSec++=*(unsigned char*)vc++;
			*bytesOfSec++=*(unsigned char*)vc++;
			*bytesOfSec++=*(unsigned char*)vc++;
			bufferDirty=1;							// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0);
			}	
		}		


	return(1);									// neuladen des fat sektors, in dem oneCluster ist, nötig !!		
}


//***************************************************************************************************************
// löscht cluster chain, beginnend ab dem startCluster.
// sucht cluster, setzt inhalt usw.. abschließend noch den cluster-chain ende markierten cluster löschen.
//***************************************************************************************************************
unsigned char fat_delClusterChain(unsigned long int startCluster){
  unsigned long int nextCluster; 
  
  nextCluster=startCluster;		// tmp variable, wegen verketteter cluster..
	
  do{
	startCluster=nextCluster;
	nextCluster=fat_getNextCluster(startCluster);
	fat_setCluster(startCluster,0x00000000);  	
	}while(!((nextCluster==0xfffffff&&fat.fatType==32)||(nextCluster==0xffff&&fat.fatType==16)));	
	
	fat_writeSector(fat.currentSectorNr);

  return(0);
}


//***************************************************************************************************************
// Initialisiert die Fat(16/32) daten, wie: root directory sektor, daten sektor, fat sektor...
// siehe auch Fatgen103.pdf. ist NICHT auf performance optimiert!
// byte/sector, byte/cluster, anzahl der fats, sector/fat ... (halt alle wichtigen daten zum lesen ders datei systems!)
//*****************************************************************<**********************************************
unsigned char fat_loadFatData(unsigned long int sec){
  void *vFatSz32;      
  void *rootDir;
 									// offset,size			
	unsigned int  		rootEntCnt;		// 17,2				// umbedingt mal nachsehen !
	unsigned int  		fatSz16;		// 22,2				sectors occupied by one fat16
	unsigned long int 	fatSz32;		// 36,4				sectors occupied by one fat32

	if(0==mmc_read_sector(sec,sector)){	// lesen von fat sector und bestimmen der wichtigen berreiche 
		
		fat.secPerClust=sector[13];			// fat.secPerClust, 13 only (power of 2)
		
		fat.bytsPerSec=sector[12];			// bytesPerSec, 12 high nibble 11 low nibble
		fat.bytsPerSec=fat.bytsPerSec<<8;
		fat.bytsPerSec+=sector[11];
		
		fat.fatSec=sector[15];				// resvdSecCnt, 15 high nibble 14 low nibble
		fat.fatSec=fat.fatSec<<8;
		fat.fatSec+=sector[14];

		rootEntCnt=sector[18];				// rootEntCnt, 18 high nibble 17 low nibble
		rootEntCnt=rootEntCnt<<8;
		rootEntCnt+=sector[17];
		
		fatSz16=sector[23];					// fatSz16, 23 high nibble 22 low nibble
		fatSz16=fatSz16<<8;
		fatSz16+=sector[22];
		
		
		fat.rootDir	 = (((rootEntCnt <<5) + fat.bytsPerSec) / fat.bytsPerSec)-1;		// ist 0 bei fat 32, sonst der root dir sektor	

		if(fat.rootDir==0){									// FAT32 spezifisch (die prüfung so, ist nicht spezifikation konform !).
		  vFatSz32=&fatSz32;								// void pointer auf fatSz32 zum schreiben von 4 bytes
		  *(unsigned char*)vFatSz32++=sector[36];				// lowest byte
		  *(unsigned char*)vFatSz32++=sector[37];				// 1. higher byte
		  *(unsigned char*)vFatSz32++=sector[38];				// 2. higher byte
		  *(unsigned char*)vFatSz32=sector[39];					// high byte

		  rootDir=&fat.rootDir;							// void pointer auf fat.rootDir zum schreiben von 4 bytes.
		  *(unsigned char*)rootDir++=sector[44];				// lowest byte
		  *(unsigned char*)rootDir++=sector[45];				// 1. higher byte
		  *(unsigned char*)rootDir++=sector[46];				// 2. higher byte
		  *(unsigned char*)rootDir=sector[47];					// high byte

		  fat.dataDirSec = fat.fatSec + (fatSz32 * sector[16]);	// data sector (beginnt mit cluster 2)		
		  fat.dir=0;											// dir auf 0==root dir, sonst 1. Cluster des dir											
		  fat.dataDirSec-=2;									// zeigt auf 1. cluster			
		  fat.fatType=32;										// fat typ					
		  fat.dataDirSec+=sec;									// addiert den startsektor auf (umrechnung von absolut auf real)
		  fat.fatSec+=sec;										// addiert den startsektor auf 	
		  }

		else{												// FAT16	spezifisch
			fat.dataDirSec = fat.fatSec + (fatSz16 * sector[16]) + fat.rootDir;		// data sektor (beginnt mit cluster 2)
			fat.rootDir=fat.dataDirSec-fat.rootDir;									// root dir sektor, da nicht im datenbereich(cluster)
			fat.dataDirSec-=2;														// zeigt auf 1. cluster		
			fat.dataDirSec+=sec;													// addiert den startsektor auf (umrechnung von absolut auf real)
			fat.fatSec+=sec;														// addiert den startsektor auf 	"
			fat.rootDir+=sec;														// addiert den startsektor auf 	"
			fat.dir=0;																// dir auf '0'==root dir, sonst 1.Cluster des dir
			fat.fatType=16;															// fat typ			
			}		

		return (0);
		}


	return (1);			// sector nicht gelesen, fat nicht initialisiert!!
}


//************************************************************************************************<<***************
// int fat sucht den 1. cluster des dateisystems (fat16/32) auch VBR genannt,  
// wenn superfloppy==0 wird der MBR ausgelesen um an VBR zu kommen.
//************************************************************************************************<<***************
unsigned char fat_initfat(void){					
  void *vSecOfFirstPartition;
  unsigned long int secOfFirstPartition=0;						// ist 1. sektor der 1. partition aus dem MBR  

	if(superfloppy==0){											// ist partitioniert
		if(0==mmc_read_sector(0,sector)){			

		  vSecOfFirstPartition=&secOfFirstPartition;		  

		  *(unsigned char*)vSecOfFirstPartition++=sector[454];
		  *(unsigned char*)vSecOfFirstPartition++=sector[455];
		  *(unsigned char*)vSecOfFirstPartition++=sector[456];
		  *(unsigned char*)vSecOfFirstPartition++=sector[457];

		  return fat_loadFatData(secOfFirstPartition);			// läd fat daten aus dem 1. sektor der patition
		  }

		}

	else {
		return fat_loadFatData(secOfFirstPartition);				// ist nicht partitioniert, läd fat daten aus sektor 0
		}

	return (1);
}
#pragma warn+