 
 
 // FAT - File - System ######################################################################################################
 // **************************************************************************************************************************
 // *   Das FAT - File - System:                                                                                             *
 // *   =====================================                                                                                *
 // *   Die Files "filefat.h" & "filefat.c"  sowie  "sdlib.h" & "sdlib.c" sind eine Portierung der mmc-0.4.4 library         *
 // *   von Daniel Platte (die mit WinAVR erstellt wurden) fr den CodeVisionAVR C-Compiler.                                 *
 // *   Hiermit ist es mglich in CodeVision eine SD-Karte zu verwenden.                                                     *
 // *   Ins Programm mssen nur die oben genannten HEADER-Files eingebunden werden, dann kann der Code, wie unten             *
 // *   erlutert, genutzt werden.                                                                                            *
 // *   [Disclaimer s. Dateiende]                                                                                                                     *
 // *                                                                                                                        *
 // **************************************************************************************************************************
 // *    Das FAT - File - System / Benutzung:                                                                                *
 // *    ====================================                                                                                *
 // *    Zunchst muss die SD-Karte initialisiert werden:                                                                    *
 // *            char = mmc_init();                      Rckgabewert 0 = OK / 1 = Fehler                                    *
 // *                                                                                                                        *
 // *    Jetzt muss das FAT-File-System initialiesiert werden:                                                               *
 // *            char =  fat_initfat();                  Rckgabewert 0 = OK / 1 = Fehler                                    *
 // *                                                                                                                        *
 // *    -----------------------------------------------------------------------------------                                 *
 // *    DATEIENOPERATIONEN                                                                                                  *
 // *    -----------------------------------------------------------------------------------                                 *                                                                                   *
 // *    Jede Aktion auf eine Datei kann nur erfolgen, wenn diese mit einer bestimmten Option                                *
 // *    geffnet wurde:                                                                                                     *
 // *            Datei erstellen & schreiben:                                                                                *
 // *                    char = ffopen('w',"DATEINAME");    Rckgabewert 0 = OK / 1 = Datei bereits vorhanden                *
 // *                                                                                 (kein Schreiben mglich)               *
 // *            Datei ffnen & Text anhngen:                                                                               *
 // *                    char = ffopen('a',"DATEINAME");    Rckgabewert 0 = OK / 1 = Datei nicht vorhanden                  *
 // *                                                                                 (kein Schreiben mglich)               *
 // *            Datei ffnen & lesen:                                                                                       *
 // *                    char = ffopen('r',"DATEINAME");    Rckgabewert 0 = OK / 1 = Datei nicht vorhanden                  *
 // *                                                                                 (kein Lesen mglich)                   *
 // *            (DATEINAME muss ein 8.3 Format haben -> z.B. "NEU     TXT")                                                 *
 // *            (                                             12345678123 )                                                 *
 // *                                                                                                                        *
 // *    Je nach ffnungsoption kann man nun:                                                                                *
 // *    In die Datei schreiben:                                                                                             *
 // *            ffwrite(text[i]);                          (z.B.: unsigned char text[9]="DU DA!\n\r";)                      *
 // *                                                                                                                        *
 // *    Aus der Datei lesen:                                                                                                *
 // *            char = ffread();                           (ffread liest ein Byte und setzt den "Cursor")                   *
 // *                                                       (automatisch auf das nchste Zeichen         )                   *
 // *                                                                                                                        *
 // *    Offset fr Leseoperation:                          Rckgabewert 0 = OK (Offset eingestellt)                         *
 // *            char = ffseek("OFFSET");                              / 1 = Fehler (Datei nicht vorhanden)                  *
 // *                                                                                                                        *
 // *    Ist die gewnsche Aktion beendet, muss die Datei wieder geschlossen werden. Dies geschieht mit:                     *
 // *            ffclose();                                                                                                  *
 // *                                                                                                                        *
 // *    Dateien knnen auch gelscht werden:                                                                                *
 // *           char = ffrm("DATEINAME");                    Rckgabewert 0 = OK (Datei gelscht)                            *
 // *                                                                   / 1 = Fehler                                         *
 // *                                                        (Die Datei muss geschlossen sein ( ffclose(); ))                *
 // *    ------------------------------------------------------------------------------------------                          *
 // *    ORDNEROPERATIONEN                                                                                                   *
 // *    ------------------------------------------------------------------------------------------                          *
 // *    Vorweg: Bei allen Ordneroperationen darf keine Datei geffnet sein!                                                 *
 // *            Ordnernamen men 11 Zeichen lang sein (auch Leerzeichen)                                                   *
 // *            Bsp.: "ORDNER1    "                                                                                         *
 // *                                                                                                                        *
 // *    Ordner im aktuellen Verzeichnis erstellen:          Rckgabewert 0 = OK (Ordner erstellt)                           *
 // *            char = ffmkdir("ORDNERNAME");                          / 1 = Fehler (Ordner bereits vorhanden)              *
 // *                                                                                                                        *
 // *    Ordner im aktuellen Verzeichnis lschen:            Rckgabewert 0 = OK (Ordner gelscht)                           *
 // *            char = ffrm("DATEINAME");                              / 1 = Fehler (Ordner nicht vorhanden)                *
 // *                                                                                                                        *
 // *    Verzeichnis wechseln (vorwrts):                    Rckgabewert 0 = OK (Verzeichnis gewechselt)                    *
 // *            char = ffcd("ORDNERNAME");                             / 1 = Fehler (Ordner nicht vorhanden)                *
 // *                                                        (Es kann nur in die jeweils nchste Ebene gewechselt werden.)   *
 // *                                                        (Springen  la "cd dir1/dir2" ist nicht mglich!            )   *
 // *                                                                                                                        *
 // *    Verzeichnis wechseln (eine Ebene rckwrts):        Rckgabewert 0 = OK (Verzeichnis gewechselt)                    *
 // *            char = ffcdLower();                                    / 1 = Fehler (hchste Ebene erreicht)                *
 // *                                                                                                                        *
 // *    Inhalt des aktuellen Verzeichnisses anzeigen:       Rckgabewert 0 = OK (Inhalt wurde angezeigt)                    *
 // *            char = ffls();                              (In der Funktion "lsRowsOfClust(...);" kann die Ausgabe)        *
 // *                                                        (beliebig angepat werden. z.B. UART,LCD,usw..         )        *  
 // *                                                                                                                        *
 // **************************************************************************************************************************                                                                   


 //-------------------------------------------------------------------------------------------------   
 //      DEFINES
 //-------------------------------------------------------------------------------------------------
  //-----FAT------------------
    #define superfloppy 1	      // wenn 0 ist partitioniert, wenn 1 ist unpartitioniert (superfloppy) !!


  //-----File-----------------
    #define ccSize 5						// größe des puffers für die cluster-chain  (vorsicht, sind long int)
    #define smallFileSys	1	    // wenn 0 dann ist kleines fat system !

 
 
 
 //-------------------------------------------------------------------------------------------------   
 //      PROTOTYPEN
 //-------------------------------------------------------------------------------------------------
  #pragma used+                                                             //verhindert Warnmeldung bei ungenutzen Funktionen
  //-----FAT-------------------  
    //--------Anwender---------
      unsigned char 	  fat_initfat(void);					                        // initalisierung (durchsucht MBR oder nicht) 
      unsigned long int fat_clustToSec(unsigned long int);			            // rechnet cluster zu 1. sektor des clusters um
      unsigned char 	  fat_writeSector(unsigned long int sec);		          // schreibt sektor auf karte
      unsigned long int fat_getNextCluster(unsigned long int oneCluster);	  // fat auf nächsten, verketteten cluster durchsuchen 
      unsigned long int fat_getFreeCluster(unsigned long int offsetCluster);// sucht leeren fat cluster aus der fat
      unsigned char	    fat_setCluster(unsigned long int cluster, unsigned long int content); // setzt cluster inhalt in der fat
      unsigned char 	  fat_delClusterChain(unsigned long int startCluster);// löscht cluster-chain in der fat
      unsigned char 	  fat_getFreeRowOfDir(unsigned long int dir);		      // durchsucht directory nach feiem eintrag
      unsigned char 	  fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);
      unsigned char 	  fat_loadSector(unsigned long int sec);			        // läd übergebenen absoluten sektor
      unsigned char 	  fat_loadFileDataFromDir(unsigned char name[]);		  // durchsucht das aktuelle directory 
      unsigned char	    fat_cd(unsigned char *);				                    // wechselt directory (start im rootDir)

    //---------Intern-----------
      unsigned char 		fat_loadFatData(unsigned long int);		              // läd fat daten 
      void 				      fat_markSector00(void);			                        // markiert puffer:sector mit 0x00
      
      // fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector  "eintrag gemacht !!"
      unsigned char     fat_getFreeRowOfCluster(unsigned long secStart);    // durchsucht cluster nach freiem eintrag
      unsigned char 		fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);
      
      // fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_cd   "daten chain"
      unsigned char 		fat_loadRowOfSector(unsigned int);		              // läd reihe des geladen sektors auf struct:file
      unsigned char 		fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]);// durchsucht die reihen des geladenen sektors 
            
                                
  //-----File-------------------
    //---------Anwender---------
      unsigned char     ffopen(unsigned char,unsigned char *);              // kann immer nur 1 datei bearbeiten.
      unsigned char     ffclose(void);						                          // muss aufgerufen werden bevor neue datei bearbeitet wird.
      unsigned char     ffseek(unsigned long int);			                    // setzt zeiger:bytesOfSec auf position in der geöffneten datei.
      unsigned char     ffcd(unsigned char *);				                      // wechselt direktory
      unsigned char     ffread(void);						                            // liest byte-weise aus der datei (puffert immer 512 bytes zwischen) 
      unsigned char     ffwrite(unsigned char);
      unsigned char     ffls(void);						                              // zeigt direktory inhalt an
      unsigned char     ffcdLower(void);					                          // geht ein direktory zurück, also cd.. (parent direktory)
      unsigned char     ffrm(unsigned char name[]);		                      // löscht datei aus aktuellem verzeichniss.      
      unsigned char     ffmkdir(unsigned char name[]);		                  // legt ordner in aktuellem verzeichniss an.
      
    //---------Intern-----------
      unsigned char 	  bufferClusterChain(unsigned long int);		          // puffert cluster-chain cluster zwischen auf puffer:clusterChain      
      void 		          lsRowsOfClust (unsigned long int start_sec);        // zeigt reihen eines clusters an, ab start_sec#
      void 		          delFile(void); 						                          // löscht datei einträge (lang und kurz) und cluster-chain
      

  #pragma used-

/*#######################################################################################
Dieses Programm ist freie Software. Sie k�nnen es unter den Bedingungen der 
GNU General Public License, wie von der Free Software Foundation ver�ffentlicht, 
weitergeben und/oder modifizieren, entweder gem�� Version 2 der Lizenz oder 
(nach Ihrer Option) jeder sp�teren Version. 

Die Ver�ffentlichung dieses Programms erfolgt in der Hoffnung, 
da� es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, 
sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT 
F�R EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License. 

Sie sollten eine Kopie der GNU General Public License zusammen mit diesem 
Programm erhalten haben. 
Falls nicht, schreiben Sie an die Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
#######################################################################################*/

